import mock
import pytest

from uaclient.update_contract_info import validate_release_series

M_PATH = "uaclient.update_contract_info."


class TestValidateReleaseSeries:
    @pytest.mark.parametrize("allowed_series", ("bionic", "jammy"))
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="jammy"),
    )
    @mock.patch("uaclient.system.get_distro_info")
    @mock.patch(M_PATH + "detach")
    @mock.patch(
        M_PATH + "_is_attached",
        return_value=mock.MagicMock(is_attached=True),
    )
    def test_validate_release_series(
        self,
        m_is_attached,
        m_detach,
        m_get_distro_info,
        m_get_release_info,
        allowed_series,
        FakeConfig,
        fake_machine_token_file,
    ):
        fake_machine_token_file._entitlements = {
            "support": {
                "entitlement": {"affordances": {"onlySeries": allowed_series}}
            }
        }
        m_get_distro_info.side_effect = [
            mock.MagicMock(
                release="20.04",
                series_codename="Bionic Beaver",
                series="bionic",
            ),
            mock.MagicMock(
                release="22.04",
                series_codename="Jammy Jellyfish",
                series="jammy",
            ),
        ]
        validate_release_series(cfg=FakeConfig(), only_series=allowed_series)
        if allowed_series:
            if allowed_series != "jammy":
                assert m_get_distro_info.call_count == 1
                assert m_detach.call_count == 1
        else:
            assert m_get_distro_info.call_count == 0
            assert m_detach.call_count == 0



import logging

from uaclient import lock, messages, system, util
from uaclient.api.u.pro.detach.v1 import detach
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def validate_release_series(cfg: UAConfig, only_series, show_message=False):
    LOG.debug("Validating release series")
    if not _is_attached(cfg).is_attached:
        return
    current_series = system.get_release_info().series
    if only_series != current_series:
        LOG.debug(
            "Detaching due to current series being %s. only_series: %s",
            current_series,
            only_series,
        )
        lock.clear_lock_file_if_present()
        detach()
        allowed_release = system.get_distro_info(only_series)
        message = messages.PRO_ONLY_ALLOWED_FOR_RELEASE.format(
            release=allowed_release.release,
            series_codename=allowed_release.series_codename,
        )
        if show_message:
            print(message)
        LOG.warning(message)



import logging

import mock
import pytest

from uaclient.upgrade_lts_contract import process_contract_delta_after_apt_lock


@pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
class TestUpgradeLTSContract:
    @mock.patch("uaclient.upgrade_lts_contract._is_attached")
    def test_unattached_noops(
        self, m_is_attached, capsys, caplog_text, FakeConfig
    ):
        m_is_attached.return_value = mock.MagicMock(is_attached=False)
        expected_logs = [
            "Check whether to upgrade-lts-contract",
            "Skipping upgrade-lts-contract. Machine is unattached",
        ]

        process_contract_delta_after_apt_lock(FakeConfig())

        assert 1 == m_is_attached.call_count
        out, _err = capsys.readouterr()
        assert "" == out
        debug_logs = caplog_text()
        for log in expected_logs:
            assert log in debug_logs

    @mock.patch("uaclient.upgrade_lts_contract._is_attached")
    @mock.patch("uaclient.upgrade_lts_contract.system.get_release_info")
    @mock.patch("uaclient.upgrade_lts_contract.system.subp")
    def test_upgrade_cancel_when_past_version_not_supported(
        self,
        m_subp,
        m_get_release_info,
        m_is_attached,
        capsys,
        caplog_text,
        FakeConfig,
    ):
        m_is_attached.return_value = mock.MagicMock(is_attached=True)
        m_get_release_info.return_value = mock.MagicMock(series="groovy")
        m_subp.return_value = ("", "")

        expected_msgs = [
            "Could not find past release for groovy",
        ]
        expected_logs = [
            "Could not find past release for groovy",
            "Check whether to upgrade-lts-contract",
        ]
        with pytest.raises(SystemExit) as execinfo:
            process_contract_delta_after_apt_lock(FakeConfig())

        assert 1 == execinfo.value.code
        assert 1 == m_is_attached.call_count
        assert 1 == m_get_release_info.call_count
        assert 1 == m_subp.call_count
        out, _err = capsys.readouterr()
        assert out == "\n".join(expected_msgs) + "\n"
        debug_logs = caplog_text()
        for log in expected_logs:
            assert log in debug_logs

    @mock.patch("uaclient.upgrade_lts_contract._is_attached")
    @mock.patch("uaclient.upgrade_lts_contract.system.get_release_info")
    @mock.patch("uaclient.upgrade_lts_contract.system.subp")
    @mock.patch(
        "uaclient.upgrade_lts_contract.contract.process_entitlements_delta"
    )
    @mock.patch("uaclient.upgrade_lts_contract.time.sleep")
    def test_upgrade_contract_when_apt_lock_is_held(
        self,
        m_sleep,
        m_process_delta,
        m_subp,
        m_get_release_info,
        m_is_attached,
        capsys,
        caplog_text,
        FakeConfig,
        fake_machine_token_file,
    ):
        m_is_attached.return_value = mock.MagicMock(is_attached=True)
        m_get_release_info.return_value = mock.MagicMock(series="focal")
        fake_machine_token_file.attached = True

        m_subp.side_effect = [
            ("apt     146195 root", ""),
            ("apt     146195 root", ""),
            ("apt     146195 root", ""),
            ("", ""),
        ]

        m_process_delta.return_value = True
        m_sleep.return_value = True

        expected_stdout = "\n".join(
            [
                "APT lock is held. Ubuntu Pro configuration will wait until "
                "it is released",
                "Starting upgrade of Ubuntu Pro service configuration",
                "Finished upgrade of Ubuntu Pro service configuration",
                "",
            ]
        )

        with mock.patch(
            "uaclient.upgrade_lts_contract.UAConfig",
            return_value=FakeConfig(),
        ):
            process_contract_delta_after_apt_lock(FakeConfig())

        assert 1 == m_is_attached.call_count
        assert 1 == m_get_release_info.call_count
        assert 4 == m_subp.call_count
        assert 1 == m_process_delta.call_count
        out, _err = capsys.readouterr()
        assert out == expected_stdout



#!/usr/bin/env python3

"""
This function is called from lib/upgrade_lts_contract.py and from
lib/reboot_cmds.py

This function should be used after running do-release-upgrade in a machine.
It will detect any contract deltas between the release before
do-release-upgrade and the current release. If we find any differences in
the uaclient contract between those releases, we will apply that difference
in the upgraded release.

For example, suppose we are on Trusty and we are upgrading to Xenial. We found
that the apt url for esm services on trusty:

https://esm.ubuntu.com/ubuntu

While on Xenial, the apt url is:

https://esm.ubuntu.com/infra/ubuntu

This script will detect differences like that and update the Xenial system
to reflect them.
"""

import logging
import sys
import time

from uaclient import contract, defaults, messages, system, util
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.files import machine_token

# We consider the past release for LTSs to be the last LTS,
# because we don't have any services available on non-LTS.
# This makes it safer for us to try to process contract deltas.
# For example, we had "jammy": "focal" even when Impish was
# still supported.
current_codename_to_past_codename = {
    "xenial": "trusty",
    "bionic": "xenial",
    "focal": "bionic",
    "jammy": "focal",
    "noble": "jammy",
    "oracular": "noble",
}

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def process_contract_delta_after_apt_lock(cfg: UAConfig) -> None:
    LOG.debug("Check whether to upgrade-lts-contract")
    if not _is_attached(cfg).is_attached:
        LOG.debug("Skipping upgrade-lts-contract. Machine is unattached")
        return
    LOG.debug("Starting upgrade-lts-contract.")
    out, _err = system.subp(["lsof", "/var/lib/apt/lists/lock"], rcs=[0, 1])
    if out:
        print(messages.RELEASE_UPGRADE_APT_LOCK_HELD_WILL_WAIT)

    current_release = system.get_release_info().series
    machine_token_file = machine_token.get_machine_token_file(cfg)

    past_release = current_codename_to_past_codename.get(current_release)
    if past_release is None:
        print(
            messages.RELEASE_UPGRADE_NO_PAST_RELEASE.format(
                release=current_release
            )
        )
        LOG.warning(
            "Could not find past release for %s. Current known releases: %r.",
            current_release,
            current_codename_to_past_codename,
        )
        sys.exit(1)

    past_entitlements = machine_token_file.entitlements(series=past_release)
    new_entitlements = machine_token_file.entitlements()

    retry_count = 0
    while out:
        # Loop until apt hold is released at the end of `do-release-upgrade`
        LOG.debug("Detected that apt lock is held. Sleeping 10 seconds.")
        time.sleep(10)
        out, _err = system.subp(
            ["lsof", "/var/lib/apt/lists/lock"], rcs=[0, 1]
        )
        retry_count += 1

    LOG.debug(
        "upgrade-lts-contract processing contract deltas: %s -> %s",
        past_release,
        current_release,
    )
    print(messages.RELEASE_UPGRADE_STARTING)

    contract.process_entitlements_delta(
        cfg=cfg,
        past_entitlements=past_entitlements,
        new_entitlements=new_entitlements,
        allow_enable=True,
        series_overrides=False,
    )
    LOG.debug("upgrade-lts-contract succeeded after %s retries", retry_count)
    print(messages.RELEASE_UPGRADE_SUCCESS)


def remove_private_esm_apt_cache():
    system.ensure_folder_absent(defaults.ESM_APT_ROOTDIR)
